using System;
using System.Collections.Generic;
using System.Text;

#if NUNIT
using NUnit.Framework;
#endif // NUNIT

using System.Net.Sockets; //eg ProtocolFamily
using System.Net;   //eg EndPoint
using InTheHand.Net.Sockets;
using InTheHand.Net;

using System.IO;


namespace Brecham.Obex.Net
{

    /// <summary>
    /// The base class of a set of classes that handle connection to a service on a peer
    /// machine.  As well as connecting, they also handle disconnect and clean-up using 
    /// the IDisposable pattern i.e. with the Dispose method.
    /// </summary>
    public abstract class TransportConnection : IDisposable
    {
        //--------------------------------------------------------------
        //--------------------------------------------------------------
       public const ProtocolFamily ProtocolFamily_Bluetooth = (ProtocolFamily)0x20;

        //--------------------------------------------------------------
        //--------------------------------------------------------------
        // The service name eg "obex-push", "obex-ftp".
        protected String m_service;
        //
        bool m_connectedTransport;
        Socket m_socket;
        NetworkStream m_peerStream;

        //--------------------------------------------------------------
        //--------------------------------------------------------------
        public Stream PeerStream
        {
            get
            {
                if (!m_connectedTransport)
                    throw new InvalidOperationException("Not connected.");
                if (m_peerStream == null) {
                    m_peerStream = new NetworkStream(m_socket, true);
                }
                return m_peerStream;
            }
        }
        
        //--------------------------------------------------------------
        //--------------------------------------------------------------
        protected TransportConnection()
        { }

        //--------------------------------------------------------------
        //--------------------------------------------------------------
        /// <summary>
        /// Returns one of "obex-push", "obex-ftp", etc.
        /// </summary>
        /// <returns></returns>
        protected abstract String ChooseService();

        /// <summary>
        /// Which L3 protocol
        /// </summary>
        /// <returns></returns>
        protected abstract ProtocolFamily ChooseProtocol();

        /// <summary>
        /// Returns the address of the peer device to connect to.  Is returned as a
        /// <c>String</c> since there&#x2019;s no <c>NetworkAddress</c> base class.
        /// </summary>
        /// <returns></returns>
        protected abstract String ChoosePeer(ProtocolFamily proto);

        /// <summary>
        /// Called to display status text to the user.  For a GUI application this
        /// would likely be overridden to set the status bar text, for a console
        /// application it would likely be overridden to simply Console.WriteLine
        /// the message.
        /// </summary>
        /// <param name="message">The status message to be displayed to the user.</param>
        protected abstract void ShowStatus(String message);

        //--------------------------------------------------------------
        //--------------------------------------------------------------
        protected abstract String GetServiceString(ProtocolFamily proto, String service);

        protected abstract Guid GetServiceGuid(ProtocolFamily proto, String service);


        //--------------------------------------------------------------
        //--------------------------------------------------------------
        protected virtual EndPoint CreateEndPoint(ProtocolFamily protocol, String addressString, String service)
        {
            EndPoint ep;
            switch (protocol) {
                case ProtocolFamily.Irda:
                    IrDAAddress addrIrda = IrDAAddress.Parse(addressString);
                    ep = new IrDAEndPoint(addrIrda, GetServiceString(protocol, service));
                    break;
                case ProtocolFamily_Bluetooth:
                    BluetoothAddress addrBt = BluetoothAddress.Parse(addressString);
                    ep = new BluetoothEndPoint(addrBt, GetServiceGuid(protocol, service));
                    break;
                default:
                    throw new ArgumentException("Unhandled protocol.");
            }//switch
            return ep;
        }

        protected  virtual Socket CreateSocket(ProtocolFamily protocol)
        {
            Socket sock;
            AddressFamily af = (AddressFamily)protocol;
            switch (protocol) {
                case ProtocolFamily.Irda:
                    sock = new Socket(af, SocketType.Stream, ProtocolType.Unspecified);
                    break;
                case ProtocolFamily_Bluetooth:
                    sock = new Socket(af, SocketType.Stream, (ProtocolType)3);
                    break;
                default:
                    throw new ArgumentException("Unhandled protocol.");
            }//switch
            return sock;
        }

        public virtual bool ConnectTransport()
        {
            ProtocolFamily proto = ChooseProtocol();
            if (proto == ProtocolFamily.Unspecified) {
                return false;
            }
            String addrString = ChoosePeer(proto);
            if (String.IsNullOrEmpty(addrString)) {
                return false;
            }
            if (addrString == null) {
                ShowStatus("No peer device chosen.");
                return false;
            } else {
                m_service = ChooseService();
                if (String.IsNullOrEmpty(m_service)) {
                    return false;
                }
                m_socket = CreateSocket(proto);
                EndPoint ep = CreateEndPoint(proto, addrString, m_service);
                ShowStatus("Connecting...");
                m_socket.Connect(ep);
                m_connectedTransport = true;
                ShowStatus("Connected.");
                return true;
            }
        }

        public virtual bool Connect()
        {
            return ConnectTransport();
        }

        //--------------------------------------------------------------
        //--------------------------------------------------------------
        public void Close()
        {
            Dispose();
        }

        #region IDisposable Members

        public void Dispose()
        {
            Dispose(true);
        }

        public virtual void Dispose(bool disposing)
        {
            System.Diagnostics.Debug.Assert(disposing);
            if (disposing) {
                if (m_socket != null) { m_socket.Close(); }
                m_peerStream = null;
            }//if
        }

        #endregion
    }//class--TransportConnection


    public abstract class ObexTransportConnection : TransportConnection
    {
        //--------------------------------------------------------------
        public const String ObexService = "obex";
        public const String ObexPushService = "obex-push";
        public const String ObexFtpService = "obex-ftp";
        protected static readonly String[] s_schemes = { ObexService, ObexPushService, ObexFtpService  /*, "obex-sync"*/ };

        //--------------------------------------------------------------
        protected ObexTransportConnection() 
            : base() 
        { }

        //--------------------------------------------------------------
        protected override string GetServiceString(ProtocolFamily proto, String service)
        {
            if (proto == ProtocolFamily.Irda) {
                return "OBEX";  // For every standard OBEX service/application.
            } else {
                throw new ArgumentException("Unhandled service type.");
            }
        }

        protected override Guid GetServiceGuid(ProtocolFamily proto, String service)
        {
            if (proto == TransportConnection.ProtocolFamily_Bluetooth) {
                switch (service) {
                    case ObexPushService: return InTheHand.Net.Bluetooth.BluetoothService.ObexObjectPush;
                    case ObexFtpService: return InTheHand.Net.Bluetooth.BluetoothService.ObexFileTransfer;
                    default: throw new ArgumentException("Unknown service type.", "service");
                }//switch

            } else {
                    throw new ArgumentException("Unhandled L3 protocol.");
            }
        }

    }//class--ObexTransportConnection 


    public abstract class ObexSessionConnection : ObexTransportConnection
    {
        //--------------------------------------------------------------
        private int m_mru = UInt16.MaxValue;
        bool m_connectedSession;
        ObexClientSession m_sess;

        //--------------------------------------------------------------
        /// <summary>
        /// Gets or set the size of buffer that the ObexClientSession will be instantiated with.
        /// That is, as passed to its constructor&#x2019;s <c>bufferSize</c> parameter.
        /// </summary>
        public int ObexBufferSize
        {
            get { return m_mru; }
            set { m_mru = value; }
        }

        public ObexClientSession ObexClientSession
        {
            get
            {
                if (!m_connectedSession)
                    throw new InvalidOperationException("Not connected.");
                return m_sess;
            }
        }

        //--------------------------------------------------------------
        protected ObexSessionConnection()
            : base()
        { }

        //--------------------------------------------------------------
        public override bool Connect()
        {
            if (!base.Connect()) {
                return false;
            }
            //
            ShowStatus("OBEX Connecting...");
            m_sess = new ObexClientSession(PeerStream, m_mru);
            byte[] target = GetObexTarget(m_service);
            if (target == null) {
                m_sess.Connect();
            } else {
                m_sess.Connect(target);
            }
            ShowStatus("OBEX Connected.");
            m_connectedSession=true;
            return true;
        }

        public virtual byte[] GetObexTarget(String service)
        {
            byte[]target;
            switch (service) {
                case ObexPushService:
                    target = null;
                    break;
                case ObexFtpService:
                    target = Brecham.Obex.ObexConstant.Target.FolderBrowsing;
                    break;
                default:
                    throw new ArgumentException("Unhandled service type >" + service + "<.");
            }
            return target;
        }
    }//class


    public class GuiObexSessionConnection : ObexSessionConnection
    {
        //--------------------------------------------------------------
        bool m_toBluetooth;
        bool m_fbService;
        System.Windows.Forms.Control m_statusControl;
#if ! NETCF
        System.Windows.Forms.ToolStripItem m_toolStripItem;
#endif // ! NETCF

        //--------------------------------------------------------------
        public GuiObexSessionConnection(bool toBluetooth, bool fbService, 
            System.Windows.Forms.Control statusControl)
            :this(toBluetooth, fbService)
        {
            m_statusControl = statusControl;
        }

#if ! NETCF
        public GuiObexSessionConnection(bool toBluetooth, bool fbService,
            System.Windows.Forms.ToolStripItem toolStripItem)
            : this(toBluetooth, fbService)
        {
            m_toolStripItem = toolStripItem;
        }
#endif // ! NETCF

        public GuiObexSessionConnection(bool toBluetooth, bool fbService)
        {
            m_toBluetooth = toBluetooth;
            m_fbService = fbService;
        }

        //--------------------------------------------------------------
        protected override void ShowStatus(string message)
        {
            if (m_statusControl != null) {
                m_statusControl.Text = message;
            }
#if ! NETCF
            if (m_toolStripItem != null) {
                m_toolStripItem.Text = message;
            }
#endif // ! NETCF
        }

        protected override string ChooseService()
        {
            if (m_fbService) {
                return ObexFtpService;
            } else {
                return ObexPushService;
            }
        }

        protected override ProtocolFamily ChooseProtocol()
        {
            if (m_toBluetooth) {
                return ProtocolFamily_Bluetooth;
            } else {
                return ProtocolFamily.Irda;
            }
        }

        protected override string ChoosePeer(ProtocolFamily proto)
        {
            if (proto == ProtocolFamily_Bluetooth) {
                return ChoosePeerBluetooth();
            } else if (proto == ProtocolFamily.Irda) {
                return ChoosePeerIrda();
            } else {
                throw new ArgumentException("Unhandled protocol.");
            }
        }

        private string ChoosePeerIrda()
        {
            IrDAClient cli = new IrDAClient();
            IrDADeviceInfo[] devs = cli.DiscoverDevices(1);
            if (devs.Length >= 1) {
                String addrAsString = devs[0].DeviceAddress.ToString();
                return addrAsString;
            } else {
                ShowStatus("No peer IrDA devices found.");
                return null;
            }
        }

        private string ChoosePeerBluetooth()
        {
            // Bluetooth
            InTheHand.Net.Forms.SelectBluetoothDeviceDialog dialog
                = new InTheHand.Net.Forms.SelectBluetoothDeviceDialog();
            // TODO (((PutGUI--SelectBluetoothDeviceDialog() filtering...?)))
            dialog.ShowUnknown = true;//HACK showNewDevices;
            //dialog.ShowUnknown = false;
            //dialog.ShowAuthenticated = false;
            System.Windows.Forms.DialogResult rslt = dialog.ShowDialog();
            if (rslt != System.Windows.Forms.DialogResult.OK)
                return null;
            BluetoothDeviceInfo device = dialog.SelectedDevice;
            if (device == null) {
                System.Diagnostics.Debug.Fail("OK, but no device!?");
                return null;
            }
            String addr = device.DeviceAddress.ToString();
            return addr;
        }

    }//class GuiObexSessionConnection 


    public class ConsoleMenuObexSessionConnection : ObexSessionConnection
    {
        //--------------------------------------------------------------
        public ConsoleMenuObexSessionConnection()
            : base()
        { }


        //--------------------------------------------------------------
        // Should accept a list of which services to offer in the constructor and build
        // a menu here based on that list.  Here we're assuming the list has 
        protected override string ChooseService()
        {
            // Check what application to connect to, we assume Folder Browsing.
            Console.Write("Connect to the Folder-Browsing server, select No for the default inbox");
            bool fbService = MenuKeyboardInput.GetCmdLineIsYesAssumeYes();
            //
            if (fbService) {
                Console.WriteLine("Will connect to the Folder-Browsing Target.");
                return ObexFtpService;
            } else {
                Console.WriteLine("Will connect to the default Inbox Target.");
                return ObexPushService;
            }
        }

        protected override ProtocolFamily ChooseProtocol()
        {
            // Check what protocol to connect over, we assume Bluetooth.
            Console.Write("Connect to a Bluetooth peer, select No for an IrDA peer");
            bool toBluetooth = MenuKeyboardInput.GetCmdLineIsYesAssumeYes();
            //
            if (toBluetooth) {
                return ProtocolFamily_Bluetooth;
            } else {
                return ProtocolFamily.Irda;
            }
        }

        protected override string ChoosePeer(ProtocolFamily proto)
        {
            if (proto == ProtocolFamily_Bluetooth) {
                return ChoosePeerBluetooth();
            } else if (proto == ProtocolFamily.Irda) {
                return ChoosePeerIrda();
            } else {
                throw new ArgumentException("Unhandled protocol.");
            }
        }

        private string ChoosePeerIrda()
        {
            IrDAClient irdaCli = null;
            try {
                irdaCli = new IrDAClient();
                IrDADeviceInfo[] devices;
                while (true) {
                    Console.WriteLine("IrDA discovery...");
                    devices = irdaCli.DiscoverDevices();
                    for (int i = 0; i < devices.Length; ++i) {
                        IrDADeviceInfo curDev = devices[i];
                        Console.WriteLine("{0,2}: {1} [{3}]",//, last seen: {2}",
                            i, curDev.DeviceName, null//curDev.LastSeen
                            , curDev
                            );
                    }//for
#if ! ( FX1_1 || NETCF )
                    Console.Beep();
#endif
                    if (devices.Length != 0) break;
                    Console.Write("No devices discovered, hit return to try again>");
                    Console.ReadLine();
                }//while
                Console.Write("Choose which device by number");
                int selected = MenuKeyboardInput.GetIndexOfRange(devices.Length);
                IrDAAddress addr = devices[selected].DeviceAddress;
                String addrAsString = addr.ToString();
                return addrAsString;
            } finally {
                if (irdaCli != null) { irdaCli.Close(); }
            }
        }

        public string ChoosePeerBluetooth()
        {
            BluetoothClient btCli = new BluetoothClient();
            BluetoothDeviceInfo[] devices;
            while (true) {
                Console.WriteLine("Bluetooth discovery...");
                devices = btCli.DiscoverDevices();
                for (int i = 0; i < devices.Length; ++i) {
                    BluetoothDeviceInfo curDev = devices[i];
                    Console.WriteLine("{0,2}: {1}, last seen: {2} [{3}]",
                        i, curDev.DeviceName, curDev.LastSeen, curDev);
                }//for
#if ! ( FX1_1 || NETCF )
                Console.Beep();
#endif
                if (devices.Length != 0) break;
                Console.Write("No devices discovered, hit return to try again>");
                Console.ReadLine();
            }//while
            Console.Write("Choose which device by number");
            int selected = MenuKeyboardInput.GetIndexOfRange(devices.Length);
            BluetoothAddress addr = devices[selected].DeviceAddress;
            return addr.ToString();
        }

        protected override void ShowStatus(string message)
        {
            Console.WriteLine(message);
        }
    }//class--ConsoleMenuObexConnection 


    public class UriObexSessionConnection : ObexSessionConnection
    {
        //--------------------------------------------------------------
        Uri m_uri;
        ProtocolFamily m_protoFromUri;
        String m_addressFromUri;

        public UriObexSessionConnection(Uri uri)
        {
            m_uri = uri;
            VerifyAndDisectUri();
        }//.ctor

        //--------------------------------------------------------------
        public const String ExMsgUriBadScheme = "Uri Scheme invalid, not a known value.";
        public const String ExMsgUriBadPathQuery = "Uri Path and Query must be empty";
        public const String ExMsgUriBadFragment = "Uri Fragment must be empty";

        //--------------------------------------------------------------
        void VerifyUri(Uri peer)
        {
            //-------------
            // Scheme
            //-------------
            String scheme = peer.Scheme;
            foreach (String curScheme in s_schemes) {
                if (curScheme.Equals(scheme, StringComparison.Ordinal))
                    goto GoodScheme;
            }
            throw new ArgumentException(ExMsgUriBadScheme);
        GoodScheme:

            //-------------
            // Path and Query
            //-------------
            if (!peer.PathAndQuery.Equals("/", StringComparison.Ordinal))
                throw new ArgumentException(ExMsgUriBadPathQuery);
            if (!String.IsNullOrEmpty(peer.Fragment))
                throw new ArgumentException(ExMsgUriBadFragment);

            //-------------
            // Host
            //-------------
            // We won't do any checking of the host part here, there are various formats
            // allowed and we'll be strict at connection time.
        }


        void VerifyAndDisectUri()
        {
            VerifyUri(m_uri);
            m_service = m_uri.Scheme;
            //
            IrDAAddress irdaAddr; BluetoothAddress btAddr;
            GetAddressFromUri(m_uri, out irdaAddr, out btAddr);
            if (btAddr != null) {
                m_protoFromUri = ProtocolFamily_Bluetooth;
                m_addressFromUri = btAddr.ToString();
            } else if (irdaAddr != null) {
                m_protoFromUri = ProtocolFamily.Irda;
                m_addressFromUri = irdaAddr.ToString();
            } else {
                System.Diagnostics.Debug.Fail( "just to show that we never get here, GetAddressFromUri will have thrown.");
            }
        }


        /// <summary>
        /// Parse and convert the host portion of the request URI in to a device 
        /// address, either a IrDA address or a Bluwtooth address based on the 
        /// format.
        /// </summary>
        /// <param name="irdaAddr"></param>
        /// <param name="bluetoothAddr"></param>
        /// <exception cref="T:System.ArgumentException">If the format of the host 
        /// portion of the given Uri is not recognized.
        /// </exception>
        void GetAddressFromUri(Uri peer, out IrDAAddress irdaAddr, out BluetoothAddress bluetoothAddr)
        {
            //
            // IrDA: 
            //   xxxxxxxx, length=8
            //   xx-xx-xx-xx, length=11
            //
            // Bluetooth: 
            //   xxxxxxxxxxxx, length=12
            //   xx-xx-xx-xx-xx-xx, length=17
            //
            String hostname = peer.Host;
            if (hostname.Length == 8
                || hostname.Length == (8 + 3)) {
                // IrDA
                irdaAddr = IrDAAddress.Parse(hostname);
                bluetoothAddr = null;
                return;
            }
            //
            if (hostname.Length == 12
                || hostname.Length == (12 + 5)) {
                //Bluetooth
                bluetoothAddr = BluetoothAddress.Parse(hostname);
                irdaAddr = null;
                return;
            }
            //
            throw new ArgumentException("Unrecognized hostname format."
                + " E.g. for an IrDA peer use format xx-xx-xx-xx, as returned by IrDAAddress.ToString.");
        }//fn


        //--------------------------------------------------------------
        protected override string ChooseService()
        {
            String svc = m_uri.Scheme;
            return svc;
        }

        protected override ProtocolFamily ChooseProtocol()
        {
            return m_protoFromUri;
        }

        protected override string ChoosePeer(ProtocolFamily proto)
        {
            return m_addressFromUri;
        }

        protected override void ShowStatus(string message)
        { /*NOOP*/; }

    }//class UriObexSessionConnection


#if NUNIT
    [TestFixture]
    public class Test_UriObexConnection_VerifyUri
    {
        UriObexSessionConnection New_UriObexConnection(Uri uri)
        {
            return new UriObexSessionConnection(uri);
        }

        [Test]
        [ExpectedException(typeof(ArgumentException), UriObexSessionConnection.ExMsgUriBadScheme)]
        public void badScheme()
        {
            New_UriObexConnection(new Uri("zzobex://12345678"));
        }

        [Test]
        public void goodIrda()
        {
            New_UriObexConnection(new Uri("obex://12345678"));
        }

        //[Test]        [Ignore]
        public void goodIrdaHyphens()
        {
            New_UriObexConnection(new Uri("obex://12-34-56-78"));
        }

        [Test]
        public void goodIrdaDots()
        {
            New_UriObexConnection(new Uri("obex://12.34.56.78"));
        }

        //[Test]        [Ignore]
        public void goodIrdaColonsA()
        {
            New_UriObexConnection(new Uri("obex://12:34:56:78"));
        }

        // [Test]        [Ignore]
        //public void goodIrdaColonsBldr()
        //{
        //    UriBuilder bldr = new UriBuilder();
        //    bldr.Host = "12:34:56:78";
        //    bldr.Scheme = "obex";
        //    new UriObexConnection(bldr.Uri);
        //}

        [Test]
        public void goodIrdaPush()
        {
            New_UriObexConnection(new Uri("obex-push://12345678/"));
        }

        [Test]
        public void goodBluetooth()
        {
            New_UriObexConnection(new Uri("obex://123456789012"));
        }

        //[Test]        [Ignore]
        public void goodBluetoothHyphens()
        {
            New_UriObexConnection(new Uri("obex://12-34-56-78-90-12"));
        }

        [Test]
        public void goodBluetoothDots()
        {
            New_UriObexConnection(new Uri("obex://12.34.56.78.90.12"));
        }

        //[Test]        [Ignore]
        public void goodBluetoothColons()
        {
            New_UriObexConnection(new Uri("obex://12:34:56:78:90:12"));
        }

        [Test]
        public void goodBluetoothFtp()
        {
            New_UriObexConnection(new Uri("obex-ftp://123456789012/"));
        }

        //Later!
        //[Test]
        //public void goodBluetoothSync()
        //{
        //    New_UriObexConnection(new Uri("obex-sync://123456789012/"));
        //}

        [Test]
        [ExpectedException(typeof(ArgumentException), UriObexSessionConnection.ExMsgUriBadPathQuery)]
        public void badPath()
        {
            New_UriObexConnection(new Uri("obex://12345678/p"));
        }

        [Test]
        [ExpectedException(typeof(ArgumentException), UriObexSessionConnection.ExMsgUriBadPathQuery)]
        public void badQuery()
        {

            New_UriObexConnection(new Uri("obex://12345678/?q"));
        }

        [Test]
        [ExpectedException(typeof(ArgumentException), UriObexSessionConnection.ExMsgUriBadFragment)]
        public void badFragment()
        {

            New_UriObexConnection(new Uri("obex://12345678#f"));
        }

    }
#endif



    //==========================================================================
    public static class MenuKeyboardInput
    {
        //----------------------------------------------------------------------
        public static int GetIndexOfRange(int range)
        {
            Int32 value = -99;
            while (true) {
                Console.Write(">");
                String line = Console.ReadLine();
                // No Int32.TryParse in CF20/FX1_1.
                // bool success = Int32.TryParse(line, out value);
                bool success = false;
                try {
                    //success = false;
                    value = Int32.Parse(line);
                    success = true;
                } catch (FormatException) { } catch (OverflowException) { }
                if (!success) {
                    Console.Write("Not a number, please try again");
                } else if (value >= range || value < 0) {
                    Console.Write("The number should be in the range, 0 through {0}", range - 1);
                } else { break; }
            }//while
            return value;
        }

        //----------------------------------------------------------------------
        public static bool GetCmdLineIsYesAssumeNo()
        {
            Console.Write(" [y/N]>");
            String input = Console.ReadLine();
            return GetCmdLineIsYesAssumeNo(input);
        }

        public static bool GetCmdLineIsYesAssumeYes()
        {
            Console.Write(" [Y/n]>");
            String input = Console.ReadLine();
            return GetCmdLineIsYesAssumeYes(input);
        }

        //----
        public static bool GetCmdLineIsYesAssumeNo(String input)
        {
            return GetCmdLineIsYesBlankIs(false, input);
        }

        public static bool GetCmdLineIsYesAssumeYes(String input)
        {
            return GetCmdLineIsYesBlankIs(true, input);
        }

        public static bool GetCmdLineIsYesBlankIs(bool assume, String input)
        {
            if (input == null
                || input.Length == 0)
                return assume;
            return FirstCharAsUpper(input) == 'Y';
        }

        //----------------------------------------------------------------------


        public static char FirstCharAsUpper(String s)
        {
#if ! FX1_1 && ! NETCF
            return s.Substring(0, 1).ToUpperInvariant()[0];
#else
            return s.Substring(0, 1).ToUpper(System.Globalization.CultureInfo.InvariantCulture)[0];
#endif
        }//fn

        //----------------------------------------------------------------------
    }//class


}

